use std::fmt::Display;
use std::ops::{Neg, Sub};
use std::str::FromStr;

use getopts::Options;
use thiserror::Error;

macro_rules! the_flag {
    () => {
        "F~L~Â~G{E1nb38V9a3V5e86HzlY27UFAErIP5mA34kES9fbu}"
    };
}

const T1337: Timestamp = Timestamp {
    hour: 13,
    minute: 37,
};
const TZERO: Timestamp = Timestamp { hour: 0, minute: 0 };

fn options() -> Options {
    let mut opts = Options::new();

    opts.reqopt(
        "f",
        "from",
        "calculate time starting from this time",
        "HH:MM",
    );
    opts.reqopt("t", "to", "calculate time until this time", "HH:MM");
    opts.optopt(
        "p",
        "pause",
        "remove this much time from the result",
        "HH:MM",
    );

    opts.optflag("h", "help", "print this help menu");

    opts
}

#[derive(PartialEq, Copy, Clone, Debug)]
struct Timestamp {
    hour: i8,
    minute: u8,
}

#[derive(Error, Debug)]
enum TimestampParseError {
    #[error("Too many or too few separators ':'")]
    BadAmountOfSeparators,
    #[error("Hours h must fullfill 0 < h < 24 (was {0})")]
    InvalidHour(i8),
    #[error("Minutes m must fullfill 0 < m < 60 (was {0})")]
    InvalidMinute(u8),
    #[error(transparent)]
    ParseInt(#[from] std::num::ParseIntError),
}

impl Sub for Timestamp {
    type Output = Timestamp;

    fn sub(self, rhs: Self) -> Self::Output {
        let t = if self.minute >= rhs.minute {
            Timestamp {
                hour: self.hour - rhs.hour,
                minute: self.minute - rhs.minute,
            }
        } else {
            Timestamp {
                hour: self.hour - rhs.hour - 1,
                minute: 60 + self.minute - rhs.minute,
            }
        };
        #[cfg(debug_assertions)]
        {
            println!("{self} - {rhs} = {t}")
        }
        t
    }
}

impl Neg for Timestamp {
    type Output = Self;

    fn neg(mut self) -> Self::Output {
        self.hour = -self.hour;
        self
    }
}

impl Display for Timestamp {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{:02}:{:02}", self.hour, self.minute)
    }
}

impl FromStr for Timestamp {
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let parts: Vec<&str> = s.split(":").collect();
        if parts.len() != 2 {
            return Err(TimestampParseError::BadAmountOfSeparators);
        }

        let hour: i8 = parts[0].parse()?;
        let minute: u8 = parts[1].parse()?;

        if hour > 59 {
            return Err(TimestampParseError::InvalidHour(hour));
        }
        if minute > 59 {
            return Err(TimestampParseError::InvalidMinute(minute));
        }

        let t = Timestamp { hour, minute };
        Ok(t)
    }

    type Err = TimestampParseError;
}

fn main() {
    let args: Vec<String> = std::env::args().collect();
    let program = args[0].clone();
    let opts = options();
    let matches = match opts.parse(&args[1..]) {
        Ok(m) => m,
        Err(e) => {
            eprintln!("{e}");
            usage(&program, &opts)
        }
    };
    let from_time: Timestamp = match matches.opt_str("from").expect("thats required").parse() {
        Ok(t) => t,
        Err(e) => {
            eprintln!("{e}");
            usage(&program, &opts)
        }
    };
    let to_time: Timestamp = match matches.opt_str("to").expect("thats required").parse() {
        Ok(t) => t,
        Err(e) => {
            eprintln!("{e}");
            usage(&program, &opts)
        }
    };
    let pause_time: Timestamp = match matches.opt_str("pause") {
        Some(s) => match s.parse() {
            Ok(t) => t,
            Err(e) => {
                eprintln!("{e}");
                usage(&program, &opts)
            }
        },
        None => TZERO,
    };

    let diff = to_time - from_time - pause_time;

    println!("from\t\t: {from_time}");
    println!("to\t\t: {to_time}");
    println!("pause\t\t: {pause_time}");
    println!("difference\t: {}", diff);

    if pause_time == T1337 && diff == -T1337 {
        println!("{}", obfstr::obfstr!(the_flag!()))
    }
}

fn usage(program: &str, opts: &Options) -> ! {
    let brief = format!("Usage: {} FILE [options]", program);
    print!("{}", opts.usage(&brief));
    std::process::exit(1)
}

#[cfg(test)]
mod tests {
    use super::*;
    const TONE: Timestamp = Timestamp { hour: 1, minute: 0 };
    const TTHREE: Timestamp = Timestamp { hour: 3, minute: 0 };
    #[test]
    fn test_sanity() {
        assert_eq!(format!("{}", -T1337), "-13:37");
        assert_eq!(1 - 5 - 5, -9);
        assert_eq!(TONE, TTHREE - TONE - TONE);
    }
}
